; =============================================================================
;                 Electronic Speed Controller for 12F675
; ============================================================================= 
;                 Bruce Abbott (bhabbott@paradise.net.nz)
;
; Changes:
;
; V1.0
;
; - Modified Spd400.asm to suit PIC12F675.
;

                PROCESSOR PIC12F675
                INCLUDE   <P12F675.inc>
                radix     dec

; simulate=1	; enable this for debugging with software simulator

version  = 1
revision = 0

; ==========================================================================
; Configuration is:
;   Master Clear pin is disabled (used as input)
;   Code protection is OFF
;   Watchdog timer is ON
;   Oscillator is internal RC

		__CONFIG _MCLRE_OFF&_INTRC_OSC_NOCLKOUT

; Bit definitions for the GPIO register and the TRIS register

StopBit		EQU       1               ; Turns on brake
RunBit		EQU       2               ; Turns on motor
ServoBit	EQU       0               ; Input servo pulse
BattOkBit	EQU       3               ; Input battery status
GreenLedBit	EQU       4               ; Hi if Watchdog timeout occured
RedLedBit	EQU       5               ; pulsed Hi for each bad servo pulse

TrisBits	EQU       H'FF'&~((1<<StopBit)|(1<<RunBit)|(1<< GreenLedBit)|(1<<RedLedBit))

; Bits to be set with the OPTION instruction
;   No wake up
;   No weak pullups
;   Timer 0 source internal
;   Which edge is don't care
;
;   Prescaler to watchdog, set to give a 272mS timeout (about 13 servo pulses).
;
;
OptionBits      EQU       B'11011100'

; ==========================================================================
; Algorithm constants
PWM_Period      EQU       47	; Number of 7-instr code cycles
PWM_Start       EQU       4	; Lowest cycle count
BadPulseBit     EQU       7	; 0x80 = pulse width out of range
BrakeEnableBit  EQU       6	; 0x40 = brake pulse width range
ArmPulsesReqd   EQU       25	; delay before arming (*20mS)
BRAKEDELAY      EQU       25	; delay before braking (*20mS) 
LOBATTDELAY     EQU       50	; braking delay if battery low

;===========================================================================
; Macro to create offsets for file registers in RAM
;

ByteAddress     SET	32    ; user RAM starts here

BYTE            MACRO	ByteName
ByteName        EQU	ByteAddress
ByteAddress     SET	ByteAddress+1
                ENDM

; ==========================================================================
; GP Register Definitions.  
;

                BYTE    Flags           ; various boolean flags            
                BYTE    PwmCount        ; counts down to zero during PWM output  
                BYTE    PwmWidth        ; Initial value to be loaded into PwmCount  
                BYTE    ServoCount      ; length of servo pulse 0-185 * 7uS (0.9-2.2mS)
                BYTE    HiCount         ; coarse servo pulse length (ServoCount/47)              
                BYTE    Command         ; PWM width (0-43) + BrakeEnableBit
                BYTE    ArmPulseCount   ; arming safety delay (counts down)
                BYTE    ServoPulse      ; current averaged servo pulse width
                BYTE    LastPulse       ; last averaged servo pulse width
                BYTE    BrakeDelayCount ; number of servo pulses until brake engages
                BYTE    StatCopy        ; copy of STATUS at startup

; Bits in Flags

BattLow         EQU       0
PriorServoBit   EQU       1

; Bits in Command

BrakeBit        EQU       6

; =========================================================================
; Helper for making strings of no-op instructions
;
NO_OP           MACRO     count
NO_OP_COUNT     SET       count
                WHILE     NO_OP_COUNT
                nop
NO_OP_COUNT     SET       NO_OP_COUNT - 1
                ENDW
                ENDM

; =========================================================================
; Macro to set the initial state of the motor fet.  W contains the count
; on exit. The code in this Macro uses 5 clocks.
;
DoRunBit        MACRO                          
                movf      PwmWidth, W         
                skpz                     
                bsf       GPIO, RunBit
                skpnz
                bcf       GPIO, RunBit
                ENDM

; =========================================================================
; Macro to keep up with the servo pulse duration
; The code in this Macro uses 2 clocks.
;
DoServoCount    MACRO                          
                btfsc     GPIO, ServoBit
                incf      ServoCount, F
                ENDM

; ===========================================================================
;
; Now for some code
;
; Critical instruction timings are shown in brackets eg. ;(2) = 2 microseconds 
; (assuming the PIC is driven with a 4Mhz oscillator).
;
Coldstart:      ORG     0                

                movf    STATUS,w        ; remember state of status register  
                movwf   StatCopy             		

		call	0x3ff		; get pre-programmed OSCCAL value

 		bsf	STATUS,RP0	; Register bank 1
        	movwf   OSCCAL          ; set oscillator frequency
		clrf	ANSEL		; Analog I/O off
        	bcf	STATUS,RP0      ; register bank 0
		movlw	b'00000111'      
        	movwf	CMCON		; Comparator off

                               
; ==========================================================================
; The following sequence is from the book.  It moves the
; prescaler from tmr0 to the watchdog without accidental resets.
;
                clrwdt                    
                clrf      TMR0            
                movlw     OptionBits | 7  
                option                    
                clrwdt
                movlw     OptionBits
                option
                clrwdt

                clrf      GPIO
                movlw     TrisBits
                tris      GPIO                  ; All outputs are low
                bsf       GPIO,RedLedBit        ; show that power is on
                btfss     StatCopy,NOT_TO
                bsf       GPIO,GreenLedBit      ; show if watchdog timed out

; wait 500mS for peripheral circuits to stabilize

		clrf	  PwmCount
delay1		clrf	  HiCount
delay2		clrwdt
		NO_OP     4
		decfsz	  HiCount
		goto	  delay2
		decfsz	  PwmCount
		goto      delay1
				

; Note that we can get to here with the brake on if the battery got low.  
; If we got here from power up, the brake is off.  In all cases, the motor 
; is off.  Don't leave with a low battery
;
Rearm:          clrwdt
                btfss     GPIO,BattOkBit       ; Don't leave if battery is low
                goto      Rearm                         
;
; Hunt for a whole servo pulse before looking for arm pulses
;
RaLoop1:        btfss     GPIO,BattOkBit
                goto      Rearm                         
                btfss     GPIO,ServoBit
                goto      RaLoop1
RaLoop2:        btfss     GPIO,BattOkBit
                goto      Rearm                         
                btfsc     GPIO,ServoBit
                goto      RaLoop2

                movlw     ArmPulsesReqd
                movwf     ArmPulseCount

RaLoop3:        clrf      ServoCount
                clrwdt
;
; Measure length of the next servo pulse                
;
RaLoop4:        btfss     GPIO, BattOkBit
                goto      Rearm                         
                btfss     GPIO, ServoBit       ; wait until servo pulse is high       
                goto      RaLoop4             
RaLoop5:        incf      ServoCount, F        ;(1)
                skpnz                          ;(2)       
                goto      Rearm                ;    restart if > 255
                NO_OP     1                    ;(1) loop must be 7 clocks                
                btfsc     GPIO, ServoBit       ;(1) 
                goto      RaLoop5              ;(2) until servo pulse is low

; accept arming pulses that are within brake range

                clrwdt
                movlw     128                  ; min count = 0.9mS 
                subwf     ServoCount,F                
                movlw     43                   ; max count = 1.2mS
                subwf     ServoCount,W         ; is this a legal arming pulse?
                skpnc
                goto      RaLoop1              ; No - start over
                decfsz    ArmPulseCount,F      ; Do we have enough arm pulses?
                goto      RaLoop3              ; Not yet.
;
; Play a 'beep' sound, indicating successful arming. Sound can be heard from 
; the motor, or (better) from a speaker connected to GP4.  
;
Beep:           bcf       GPIO,StopBit         ; brake off                           
                clrf      PwmCount
beep1:          bsf       GPIO,RunBit          ; motor on  
                NO_OP     6                     
                bsf       GPIO,RedLedBit       ; start speaker pulse 
                NO_OP     6                     
                bcf       GPIO,RunBit          ; motor off
                movlw     60
                movwf     PwmWidth
beep2:          clrwdt                     
                decfsz    PwmWidth,F           ; wait 0.24mS
                goto      beep2                 
                bcf       GPIO,RedLedBit       ; end speaker pulse
                movlw     60
                movwf     PwmWidth
beep3:          clrwdt                     
                decfsz    PwmWidth,F           ; wait 0.24mS
                goto      beep3                
                decfsz    PwmCount,F
                goto      beep1                ; do 256 cycles (1/4 second)
;
; wait for end of next servo pulse.
;
servo_hi:       btfss     GPIO,ServoBit        ; wait for start of pulse       
                goto      servo_hi             
servo_lo:       btfsc     GPIO,ServoBit        ; wait for end of pulse
                goto      servo_lo                              

Armed:          clrf      PwmWidth             ; Hold motor off
                clrf      ServoCount
                clrf      HiCount              
                clrf      LastPulse            
                clrf      Flags                ; Servo input low, battery OK
                movlw     BRAKEDELAY           
                movwf     BrakeDelayCount      ; Precharge brake delay 
;
; Arming complete, jump to main operating loop. The motor is off, and 
; will stay off until the next servo pulse has been processed.
;
                goto      MainPwmLoop


; ==========================================================================
; PWM pulse routine.  On entry, the motor may be on or off. If the motor is 
; off, it will never be turned on.  However, it will be turned off when (if) 
; the PWM_Count register reaches zero.
;
; Also measures servo pulse length, and checks for low battery.
; 
; Takes 305 clocks to execute, uses 259 words of program space.
;
DoPwmPulse:
PWM_LOOP_COUNT  SET       PWM_Period - PWM_Start  ; expand code to 43 iterations
                WHILE     PWM_LOOP_COUNT
                
                btfss     GPIO,BattOkBit        ;(2) check for low battery
                bsf       Flags,BattLow
                DoServoCount                    ;(2) measure servo pulse length
                decf      PwmCount,F            ;(1)
                skpnz                           ;(1) end of PWM hi pulse?
                bcf       GPIO, RunBit          ;(1) yes, then turn off motor
                
PWM_LOOP_COUNT  SET       PWM_LOOP_COUNT - 1	; next code iteration
                ENDW
                
                retlw     0                     ;(2)

; ==========================================================================
;                 Servo pulse to command converter
;
; Input:  W = index 0-184 (servo pulse width) 
;                         
; Output: W = PWM_width (0-63) + Invalid (bit7) + brake(bit6)
;
; This routine takes 9 clocks to execute.
;

ServoLookup:   	movlw	  H'03'		; set page for computed goto
		movwf	  PCLATH
                movf      ServoPulse, W ; get table index 
		movwf     PCL           ; computed goto 0x300+w 
 
; ========================================================================
; The main PWM loop
;
; Two timing loops have to be simultaneously maintained:-
;
; 1/ Measure servo pulse width every 7 clocks. DoServoCount takes 
;    2 clocks, thus 5 clocks are available for other processing.
;  
; 2/ DoRunBit + 17 clocks + DoPwmPulse + 2 clocks etc. This creates 
;    the PWM cycle time of 329 clocks, which corresponds to a PWM
;    frequency of 3Khz.
;
; NOTE: DoServoCount occurs 43 times inside DoPwmPulse!
;
ServoIsHigh:    bsf       Flags, PriorServoBit ;(1) servo pulse has started!
                DoServoCount                   ;(2) 3rd precycle
                btfsc     Flags, BattLow       ;(2) If battery was ever low, quit now
                goto      BattLoShutdown       ;    
                incf      HiCount,F            ;(1) HiCount = ServoCount/47 (approx)
                skpnz                          ;(1) Limit HiCount to 255 maximum
                decf      HiCount,F            ;(1)
MainPwmLoop:    DoServoCount                   ;(2) 4th Precycle
                NO_OP     1                    ;(1)                
                call      DoPwmPulse           ;(4,2,5) PWM,DoServoCount,PWM
                DoServoCount                   ;(2) 1st Precycle
MainPwmEntry:   DoRunBit                       ;(5)
                DoServoCount                   ;(2) 2nd precycle
                movwf     PwmCount             ;(1)
                btfsc     GPIO, ServoBit       ;(2) is servo pulse Hi ? 
                goto      ServoIsHigh          ;(+1)  
                btfsc     Flags, PriorServoBit ;(2) Servo pulse now low, but was it high?
                goto      EndOfPulse           ;(+1)yes, then it is end of pulse
                clrf      ServoCount           ;(1) no, so reset servo pulse count
                DoServoCount                   ;(2) 3rd precycle
                btfsc     Flags, BattLow       ;(2) 
                goto      BattLoShutdown       ;    If battery was ever low, quit now
                goto      MainPwmLoop          ;(2) 
                


; ============================================================================
; End of Servo Pulse. We don't have to count servo pulses now, but we still
; have to maintain the PWM cycle timing! 
;
; DoRunBit  
; +17 cycles
; DoPwmCycle 
; +2 cycles
; DoRunBit
; etc.
;
; Starting 8 clocks after last DoRunBit 
;
EndOfPulse:     movf      ServoCount,w          ;(1) save servo pulse count (it could 
                movwf     ServoPulse            ;(1) be corrupted by next DoPwmPulse)
                movlw     3                     ;(1)
                subwf     HiCount,W             ;(1) HiCount < 3 ?
                skpc                            ;(1)
                bsf       Flags,BadPulseBit     ;(1) servo pulse may be too short!
                skpc                            ;(1) 
                btfsc     ServoPulse,7          ;(1) servo count >= 128 ?
                bcf       Flags,BadPulseBit     ;(1) if yes, then pulse not too short  
                call      DoPwmPulse            ;<<< 
                NO_OP     2                     ;(2)
                DoRunBit                        ;>>>
                movwf     PwmCount              ;(1)
                movlw     8                     ;(1)
                subwf     HiCount,W             ;(1) HiCount >= 8 PWM cycles ?
                skpnc                           ;(1)
                bsf       Flags,BadPulseBit     ;(1) servo pulse is too long!
                clrf      HiCount               ;(1) reset HiCount
                movlw     128                   ;(1) min servo pulse width = 0.9mS
                subwf     ServoPulse,F          ;(1)                              
                movlw     186                   ;(1) max servo pulse width = 2.2mS
                subwf     ServoPulse,W          ;(1)  
                skpnc                           ;(1) pulse width in range ?  
                bsf       Flags,BadPulseBit     ;(1)                 
                btfsc     Flags,BadPulseBit     ;(2)  ignore bad servo pulse!
                goto      BadPulse              ;(+1)
                movf      LastPulse,w           ;(1)  calculate average of count and
                addwf     ServoPulse,F          ;(1)  last count, to reduce jitter. 
                rrf       ServoPulse,F          ;(1)  average = (count + last) / 2
                call      DoPwmPulse            ;<<<  
		NO_OP	  2			;(2)
                DoRunBit                        ;>>>  still using old PWM                                        
                movwf     PwmCount              ;(1) 
                call      ServoLookup           ;(9)  decode servo pulse length  
                movwf     Command               ;(1) 
                andlw     63                    ;(1)  remove Brake bit 
                movwf     PwmWidth              ;(1)  set new PWM width
                btfsc     Command,BrakeBit      ;(2)  Braking?
                goto      Braking               ;(+1) 

NoBrake:        bcf       GPIO, StopBit         ;(1)  Turn off the brake 
                NO_OP     1                     ;(1)
                call      DoPwmPulse            ;<<<  still using old PWM 
                movlw     BRAKEDELAY            ;(1) 
                movwf     BrakeDelayCount       ;(1)  reset brake delay count
                DoRunBit                        ;>>>
                movwf     PwmCount              ;(1)  new PWM set
                goto      GoodPulse             ;(2)
;
; Here the brake is to be turned on. Let the motor slow down first. PWM 
; timing isn't critical while braking, because the motor is always off.
;
Braking:        bcf       GPIO,RunBit           ;(1)  Stop the motor
                call      DoPwmPulse            ;<<<  For delay only 
                decfsz    BrakeDelayCount,F     ;(2)  Time to turn on the Brake?
                goto      GoodPulse             ;(+1) If not, wait for another servo pulse

BrakeOn:        incf      BrakeDelayCount,F     ;(1)  Don't let count go below zero!
                bsf       GPIO, StopBit         ;(1)  Turn on the brake
 
GoodPulse:      movf      ServoPulse,W          ;(1)  remember good servo pulse
                movwf     LastPulse             ;(1)  
                clrwdt                          ;(1)  reset the watchdog timer
NextPulse:      NO_OP     9                     ;(9)
                bcf       Flags, PriorServoBit  ;(1)  servo pulse is finished      
                clrf      ServoCount            ;(1)  reset servo pulse count
                call      DoPwmPulse            ;<<<
                goto      MainPwmEntry          ;(2)  DoRunBit is next

;
; Servo pulse had bad timing, so just keep the existing PWM going. 
; Don't remember the bad pulse. 
;
BadPulse:       bsf       GPIO,RedLedBit        ;(1) start glitch counter pulse 
                NO_OP     1                     ;(1)
                call      DoPwmPulse            ;<<< still using old PWM
                bcf       GPIO,RedLedBit        ;(1) end glitch counter pulse
                bcf       Flags,BadPulseBit     ;(1)
                DoRunBit                        ;>>> still using old PWM
                movwf     PwmCount              ;(1)
                clrwdt                          ;(1) reset the watchdog timer
                NO_OP     2                     ;(2) 
                goto      NextPulse             ;(2)  


;
; Low battery - shutdown and wait for arming.
;

BattLoShutdown: bcf       GPIO, RunBit          ; Stop the motor

; Delay to let motor slow down before turning on brake

                movlw     LOBATTDELAY           ; set low battery brake delay count 
                movwf     BrakeDelayCount       
LoBattServoLo:  clrwdt                          
                btfss     GPIO, ServoBit        ; start of next pulse?
                goto      LoBattServoLo         ; no, wait for start of pulse
LoBattServoHi:  clrwdt				; reset the watchdog timer
                btfsc     GPIO, ServoBit        ; end of pulse?
                goto      LoBattServoHi         ; no, wait for end of pulse
                decfsz    BrakeDelayCount,F     ; count down, brake delay done? 
                goto      LoBattServoLo         ; no, wait for another servo pulse

                bsf       GPIO, StopBit         ; brake on
                goto      Rearm                 


;---------------------------------------------------------------------------
;                    Throttle Translation Table
;---------------------------------------------------------------------------

		ORG	  H'300'
 
                INCLUDE   "spd475.inc"  ; table of RETLW instructions



;---------------------------------------------------------------------------
;                          version string
;---------------------------------------------------------------------------

                ORG       H'3d0'
 
                dw        'S'
                dw        'p'
                dw        'e'
                dw        'e'
                dw        'd'
                dw        '4'
                dw        '7'
                dw        '5'
                dw        '>'
                dw        '>'
                dw        'V'
                dw        version+48
                dw        '.'
                dw        revision+48
                dw        '<'
                dw        '<'
                dw        'b'
                dw        'h'
                dw        'a'
                dw        'b'
                dw        'b'
                dw        'o'
                dw        't'
                dw        't'
                dw        '@'
                dw        'p'
                dw        'a'
                dw        'r'
                dw        'a'
                dw        'd'
                dw        'i'
                dw        's'
                dw        'e'
                dw        '.' 
                dw        'n'
                dw        'e'
                dw        't'
                dw        '.'
                dw        'n'
                dw        'z'
                
		org	H'3ff'
		ifdef	simulate
		retlw	H'90'		; OSCAL routine
		endif
		
                END
                
